@@ -3,8 +3,6 @@
  *
  * @APPLE_LICENSE_HEADER_START@
  * 
- * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
- * 
  * This file contains Original Code and/or Modifications of Original Code
  * as defined in and that are subject to the Apple Public Source License
  * Version 2.0 (the 'License'). You may not use this file except in
@@ -25,8 +23,40 @@
     Change History (most recent first):
 
 $Log: JNISupport.c,v $
+Revision 1.12  2005/07/13 19:20:32  cheshire
+<rdar://problem/4175511> Race condition in Java API
+Additional cleanup suggested by Roger -- NewContext() doesn't need ownerClass parameter any more
+
+Revision 1.11  2005/07/11 01:55:21  cheshire
+<rdar://problem/4175511> Race condition in Java API
+
+Revision 1.10  2005/07/05 13:01:52  cheshire
+<rdar://problem/4169791> If mDNSResponder daemon is stopped, Java API spins, burning CPU time
+
+Revision 1.9  2004/12/11 03:01:00  rpantos
+<rdar://problem/3907498> Java DNSRecord API should be cleaned up
+
+Revision 1.8  2004/11/30 23:51:05  cheshire
+Remove double semicolons
+
+Revision 1.7  2004/11/23 08:12:04  shersche
+Implement if_nametoindex and if_indextoname for Win32 platforms
+
+Revision 1.6  2004/11/23 03:41:14  cheshire
+Change JNISupport.c to call if_indextoname & if_nametoindex directly.
+(May require some additional glue code to work on Windows.)
+
+Revision 1.5  2004/11/17 17:07:44  cheshire
+Updated comment about AUTO_CALLBACKS
+
+Revision 1.4  2004/11/12 03:23:09  rpantos
+rdar://problem/3809541 implement getIfIndexForName, getNameForIfIndex.
+
+Revision 1.3  2004/06/18 04:44:17  rpantos
+Adapt to API unification on Windows
+
 Revision 1.2  2004/05/28 23:34:42  ksekar
-<rdar://problem/3672903>: Java project doesn't build on Tiger8A132
+<rdar://problem/3672903>: Java project build errors
 
 Revision 1.1  2004/04/30 16:29:35  rpantos
 First checked in.
@@ -34,46 +64,45 @@ First checked in.
 
 	This file contains the platform support for DNSSD and related Java classes.
 	It is used to shim through to the underlying <dns_sd.h> API.
-
-	To do:
-	- normalize code to eliminate USE_WIN_API
  */
 
 // AUTO_CALLBACKS should be set to 1 if the underlying mDNS implementation fires response
-// callbacks automatically (as it does on Windows).
+// callbacks automatically (as in the early Windows prototypes).
 // AUTO_CALLBACKS should be set to 0 if the client must call DNSServiceProcessResult() to
-// invoke response callbacks (as is true on Mac OS X).
+// invoke response callbacks (as is true on Mac OS X, Posix, Windows, etc.).
+// (Invoking callbacks automatically on a different thread sounds attractive, but while
+// the client gains by not needing to add an event source to its main event loop, it loses
+// by being forced to deal with concurrency and locking, which can be a bigger burden.)
 #ifndef	AUTO_CALLBACKS
 #define	AUTO_CALLBACKS	0
 #endif
 
-// (Temporary, I hope - RNP)
-// USE_WIN_API should be set to 1 to use the DNSSD.h API (on Windows).
-// USE_WIN_API should be set to 0 to use the dns_sd.h API (on everything else).
-#ifndef	USE_WIN_API
-#define	USE_WIN_API	0
-#endif
+#if !AUTO_CALLBACKS
+#ifdef _WIN32
+#include <winsock2.h>
+#else //_WIN32
+#include <sys/types.h>
+#include <sys/select.h>
+#endif // _WIN32
+#endif // AUTO_CALLBACKS
+
+#include <dns_sd.h>
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#ifdef _WIN32
+#include <winsock2.h>
+#include <iphlpapi.h>
+static char	*	if_indextoname( DWORD ifIndex, char * nameBuff);
+static DWORD	if_nametoindex( const char * nameStr );
+#define IF_NAMESIZE MAX_ADAPTER_NAME_LENGTH
+#else // _WIN32
+#include <sys/socket.h>
+#include <net/if.h>
+#endif // _WIN32
 #include <jni.h>
 
-#if USE_WIN_API
-#include <DNSSD.h>
-#else
-#include <dns_sd.h>
-#endif
-
-#if !USE_WIN_API
-#define CALLBACK_COMPAT	
-#endif
-
-#if !AUTO_CALLBACKS
-#include <sys/types.h>
-#include <sys/select.h>
-#endif
-
 #include "DNSSD.java.h"
 
 // convenience definition 
@@ -105,23 +134,13 @@ JavaVM		*gJavaVM = NULL;
 #endif
 
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_InitLibrary( JNIEnv *pEnv, jclass cls, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_InitLibrary( JNIEnv *pEnv, jclass cls,
 						jint callerVersion)
 {
 	/* Ensure that caller & interface versions match. */
 	if ( callerVersion != kInterfaceVersion)
 		return kDNSServiceErr_Incompatible;
 
-#if USE_WIN_API
-	{
-		DNSServiceErrorType		err;
-		
-		err = DNSServiceInitialize( kDNSServiceInitializeFlagsNone, 0);
-		if ( err != kDNSServiceErr_NoError)
-			return err;
-	}
-#endif
-
 #if AUTO_CALLBACKS
 	{
 		jsize	numVMs;
@@ -185,22 +204,22 @@ static void	TeardownCallbackState( void )
 #endif	// AUTO_CALLBACKS
 
 
-static OpContext	*NewContext( JNIEnv *pEnv, jobject owner, const char *ownerClass,
+static OpContext	*NewContext( JNIEnv *pEnv, jobject owner,
 								const char *callbackName, const char *callbackSig)
 // Create and initialize a new OpContext.
 {
-	OpContext				*pContext = (OpContext*) malloc( sizeof *pContext);;
+	OpContext				*pContext = (OpContext*) malloc( sizeof *pContext);
 
 	if ( pContext != NULL)
 	{
-		jfieldID		clientField = (*pEnv)->GetFieldID( pEnv, (*pEnv)->GetObjectClass( pEnv, owner), 
-															"fClient", ownerClass);
+		jfieldID		clientField = (*pEnv)->GetFieldID( pEnv, (*pEnv)->GetObjectClass( pEnv, owner),
+															"fListener", "Lcom/apple/dnssd/BaseListener;");
 
 		pContext->JavaObj = (*pEnv)->NewWeakGlobalRef( pEnv, owner);	// must convert local ref to global to cache;
 		pContext->ClientObj = (*pEnv)->GetObjectField( pEnv, owner, clientField);
 		pContext->ClientObj = (*pEnv)->NewWeakGlobalRef( pEnv, pContext->ClientObj);	// must convert local ref to global to cache
-		pContext->Callback = (*pEnv)->GetMethodID( pEnv, 
-								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj), 
+		pContext->Callback = (*pEnv)->GetMethodID( pEnv,
+								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj),
 								callbackName, callbackSig);
 		pContext->Callback2 = NULL;		// not always used
 	}
@@ -213,7 +232,7 @@ static void			ReportError( JNIEnv *pEnv, jobject target, jobject service, DNSSer
 // Invoke operationFailed() method on target with err.
 {
 	jclass			cls = (*pEnv)->GetObjectClass( pEnv, target);
-	jmethodID		opFailed = (*pEnv)->GetMethodID( pEnv, cls, "operationFailed", 
+	jmethodID		opFailed = (*pEnv)->GetMethodID( pEnv, cls, "operationFailed",
 								"(Lcom/apple/dnssd/DNSSDService;I)V");
 
 	(*pEnv)->CallVoidMethod( pEnv, target, opFailed, service, err);
@@ -230,6 +249,7 @@ JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleService_HaltOperation( JNIEnv *
 		OpContext	*pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, pThis, contextField);
 		if ( pContext != NULL)
 		{
+			// MUST clear fNativeContext first, BEFORE calling DNSServiceRefDeallocate()
 			(*pEnv)->SetIntField( pEnv, pThis, contextField, 0);
 			if ( pContext->ServiceRef != NULL)
 				DNSServiceRefDeallocate( pContext->ServiceRef);
@@ -274,7 +294,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleService_BlockForData( JNIEnv *p
 }
 
 
-JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleService_ProcessResults( JNIEnv *pEnv, jobject pThis)
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleService_ProcessResults( JNIEnv *pEnv, jobject pThis)
 /* Call through to DNSServiceProcessResult() while data remains on socket. */
 {
 #if !AUTO_CALLBACKS	// ProcessResults() not supported with AUTO_CALLBACKS 
@@ -282,6 +302,7 @@ JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleService_ProcessResults( JNIEnv
 	jclass			cls = (*pEnv)->GetObjectClass( pEnv, pThis);
 	jfieldID		contextField = (*pEnv)->GetFieldID( pEnv, cls, "fNativeContext", "I");
 	OpContext		*pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, pThis, contextField);
+	DNSServiceErrorType err = kDNSServiceErr_BadState;
 
 	if ( pContext != NULL)
 	{
@@ -294,17 +315,27 @@ JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleService_ProcessResults( JNIEnv
 		FD_ZERO( &readFDs);
 		FD_SET( sd, &readFDs);
 
-		while ( 0 < select( sd + 1, &readFDs, (fd_set*) NULL, (fd_set*) NULL, &zeroTimeout))
+		err = kDNSServiceErr_NoError;
+		if (0 < select(sd + 1, &readFDs, (fd_set*) NULL, (fd_set*) NULL, &zeroTimeout))
 		{
-			DNSServiceProcessResult( pContext->ServiceRef);
+			err = DNSServiceProcessResult(pContext->ServiceRef);
+			// Use caution here!
+			// We cannot touch any data structures associated with this operation!
+			// The DNSServiceProcessResult() routine should have invoked our callback,
+			// and our callback could have terminated the operation with op.stop();
+			// and that means HaltOperation() will have been called, which frees pContext.
+			// Basically, from here we just have to get out without touching any stale
+			// data structures that could blow up on us! Particularly, any attempt
+			// to loop here reading more results from the file descriptor is unsafe.
 		}
 	}
+	return err;
 #endif // AUTO_CALLBACKS
 }
 
 
-static void CALLBACK_COMPAT	ServiceBrowseReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex, 
-								DNSServiceErrorType errorCode, const char *serviceName, const char *regtype, 
+static void DNSSD_API	ServiceBrowseReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex,
+								DNSServiceErrorType errorCode, const char *serviceName, const char *regtype,
 								const char *replyDomain, void *context)
 {
 	OpContext		*pContext = (OpContext*) context;
@@ -315,9 +346,9 @@ static void CALLBACK_COMPAT	ServiceBrowseReply( DNSServiceRef sdRef _UNUSED, DNS
 	{
 		if ( errorCode == kDNSServiceErr_NoError)
 		{
-			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, 
-								( flags & kDNSServiceFlagsAdd) != 0 ? pContext->Callback : pContext->Callback2, 
-								pContext->JavaObj, flags, interfaceIndex, 
+			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj,
+								( flags & kDNSServiceFlagsAdd) != 0 ? pContext->Callback : pContext->Callback2,
+								pContext->JavaObj, flags, interfaceIndex,
 								(*pContext->Env)->NewStringUTF( pContext->Env, serviceName),
 								(*pContext->Env)->NewStringUTF( pContext->Env, regtype),
 								(*pContext->Env)->NewStringUTF( pContext->Env, replyDomain));
@@ -329,7 +360,7 @@ static void CALLBACK_COMPAT	ServiceBrowseReply( DNSServiceRef sdRef _UNUSED, DNS
 	TeardownCallbackState();
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleBrowser_CreateBrowser( JNIEnv *pEnv, jobject pThis, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleBrowser_CreateBrowser( JNIEnv *pEnv, jobject pThis,
 							jint flags, jint ifIndex, jstring regType, jstring domain)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -338,7 +369,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleBrowser_CreateBrowser( JNIEnv *
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 
 	if ( contextField != 0)
-		pContext = NewContext( pEnv, pThis, "Lcom/apple/dnssd/BrowseListener;", "serviceFound", 
+		pContext = NewContext( pEnv, pThis, "serviceFound",
 								"(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
 	else
 		err = kDNSServiceErr_BadParam;
@@ -348,8 +379,8 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleBrowser_CreateBrowser( JNIEnv *
 		const char	*regStr = SafeGetUTFChars( pEnv, regType);
 		const char	*domainStr = SafeGetUTFChars( pEnv, domain);
 
-		pContext->Callback2 = (*pEnv)->GetMethodID( pEnv, 
-								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj), 
+		pContext->Callback2 = (*pEnv)->GetMethodID( pEnv,
+								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj),
 								"serviceLost", "(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
 
 		err = DNSServiceBrowse( &pContext->ServiceRef, flags, ifIndex, regStr, domainStr, ServiceBrowseReply, pContext);
@@ -368,8 +399,8 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleBrowser_CreateBrowser( JNIEnv *
 }
 
 
-static void CALLBACK_COMPAT	ServiceResolveReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex, 
-								DNSServiceErrorType errorCode, const char *fullname, const char *hosttarget, 
+static void DNSSD_API	ServiceResolveReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex,
+								DNSServiceErrorType errorCode, const char *fullname, const char *hosttarget,
 								uint16_t port, uint16_t txtLen, const char *txtRecord, void *context)
 {
 	OpContext		*pContext = (OpContext*) context;
@@ -402,8 +433,8 @@ static void CALLBACK_COMPAT	ServiceResolveReply( DNSServiceRef sdRef _UNUSED, DN
 			txtObj = (*pContext->Env)->NewObject( pContext->Env, txtCls, txtCtor, txtBytes);
 			(*pContext->Env)->DeleteLocalRef( pContext->Env, txtBytes);
 	
-			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback, 
-								pContext->JavaObj, flags, interfaceIndex, 
+			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback,
+								pContext->JavaObj, flags, interfaceIndex,
 								(*pContext->Env)->NewStringUTF( pContext->Env, fullname),
 								(*pContext->Env)->NewStringUTF( pContext->Env, hosttarget),
 								port, txtObj);
@@ -415,7 +446,7 @@ static void CALLBACK_COMPAT	ServiceResolveReply( DNSServiceRef sdRef _UNUSED, DN
 	TeardownCallbackState();
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleResolver_CreateResolver( JNIEnv *pEnv, jobject pThis, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleResolver_CreateResolver( JNIEnv *pEnv, jobject pThis,
 							jint flags, jint ifIndex, jstring serviceName, jstring regType, jstring domain)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -424,7 +455,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleResolver_CreateResolver( JNIEnv
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 
 	if ( contextField != 0)
-		pContext = NewContext( pEnv, pThis, "Lcom/apple/dnssd/ResolveListener;", "serviceResolved", 
+		pContext = NewContext( pEnv, pThis, "serviceResolved",
 								"(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;Ljava/lang/String;ILcom/apple/dnssd/TXTRecord;)V");
 	else
 		err = kDNSServiceErr_BadParam;
@@ -435,7 +466,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleResolver_CreateResolver( JNIEnv
 		const char	*regStr = SafeGetUTFChars( pEnv, regType);
 		const char	*domainStr = SafeGetUTFChars( pEnv, domain);
 
-		err = DNSServiceResolve( &pContext->ServiceRef, flags, ifIndex, 
+		err = DNSServiceResolve( &pContext->ServiceRef, flags, ifIndex,
 								servStr, regStr, domainStr, ServiceResolveReply, pContext);
 		if ( err == kDNSServiceErr_NoError)
 		{
@@ -453,8 +484,8 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleResolver_CreateResolver( JNIEnv
 }
 
 
-static void CALLBACK_COMPAT	ServiceRegisterReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, 
-								DNSServiceErrorType errorCode, const char *fullname, 
+static void DNSSD_API	ServiceRegisterReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags,
+								DNSServiceErrorType errorCode, const char *serviceName,
 								const char *regType, const char *domain, void *context)
 {
 	OpContext		*pContext = (OpContext*) context;
@@ -465,9 +496,9 @@ static void CALLBACK_COMPAT	ServiceRegisterReply( DNSServiceRef sdRef _UNUSED, D
 	{
 		if ( errorCode == kDNSServiceErr_NoError)
 		{
-			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback, 
-								pContext->JavaObj, flags, 
-								(*pContext->Env)->NewStringUTF( pContext->Env, fullname),
+			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback,
+								pContext->JavaObj, flags,
+								(*pContext->Env)->NewStringUTF( pContext->Env, serviceName),
 								(*pContext->Env)->NewStringUTF( pContext->Env, regType),
 								(*pContext->Env)->NewStringUTF( pContext->Env, domain));
 		}
@@ -477,8 +508,8 @@ static void CALLBACK_COMPAT	ServiceRegisterReply( DNSServiceRef sdRef _UNUSED, D
 	TeardownCallbackState();
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_BeginRegister( JNIEnv *pEnv, jobject pThis, 
-							jint ifIndex, jint flags, jstring serviceName, jstring regType, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_BeginRegister( JNIEnv *pEnv, jobject pThis,
+							jint ifIndex, jint flags, jstring serviceName, jstring regType,
 							jstring domain, jstring host, jint port, jbyteArray txtRecord)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -489,7 +520,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_BeginRegister( JNI
 	jsize					numBytes;
 
 	if ( contextField != 0)
-		pContext = NewContext( pEnv, pThis, "Lcom/apple/dnssd/RegisterListener;", "serviceRegistered", 
+		pContext = NewContext( pEnv, pThis, "serviceRegistered",
 								"(Lcom/apple/dnssd/DNSSDRegistration;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V");
 	else
 		err = kDNSServiceErr_BadParam;
@@ -510,7 +541,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_BeginRegister( JNI
 		numBytes = txtRecord ? (*pEnv)->GetArrayLength( pEnv, txtRecord) : 0;
 
 		err = DNSServiceRegister( &pContext->ServiceRef, flags, ifIndex, servStr, regStr,  
-								domainStr, hostStr, portBits, 
+								domainStr, hostStr, portBits,
 								numBytes, pBytes, ServiceRegisterReply, pContext);
 		if ( err == kDNSServiceErr_NoError)
 		{
@@ -531,7 +562,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_BeginRegister( JNI
 	return err;
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_AddRecord( JNIEnv *pEnv, jobject pThis, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_AddRecord( JNIEnv *pEnv, jobject pThis,
 							jint flags, jint rrType, jbyteArray rData, jint ttl, jobject destObj)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -564,23 +595,28 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_AddRecord( JNIEnv
 	return err;
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_UpdateRecord( JNIEnv *pEnv, jobject pThis, 
-							jobject destObj, jint flags, jbyteArray rData, jint ttl)
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSRecord_Update( JNIEnv *pEnv, jobject pThis,
+														jint flags, jbyteArray rData, jint ttl)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
-	jfieldID				contextField = (*pEnv)->GetFieldID( pEnv, cls, "fNativeContext", "I");
-	jclass					destCls = (*pEnv)->GetObjectClass( pEnv, destObj);
-	jfieldID				recField = (*pEnv)->GetFieldID( pEnv, destCls, "fRecord", "I");
+	jfieldID				ownerField = (*pEnv)->GetFieldID( pEnv, cls, "fOwner", "Lcom/apple/dnssd/AppleService;");
+	jfieldID				recField = (*pEnv)->GetFieldID( pEnv, cls, "fRecord", "I");
 	OpContext				*pContext = NULL;
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 	jbyte					*pBytes;
 	jsize					numBytes;
 	DNSRecordRef			recRef = NULL;
 
-	if ( contextField != 0)
-		pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, pThis, contextField);
+	if ( ownerField != 0)
+	{
+		jobject		ownerObj = (*pEnv)->GetObjectField( pEnv, pThis, ownerField);
+		jclass		ownerClass = (*pEnv)->GetObjectClass( pEnv, ownerObj);
+		jfieldID	contextField = (*pEnv)->GetFieldID( pEnv, ownerClass, "fNativeContext", "I");
+		if ( contextField != 0)
+			pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, ownerObj, contextField);
+	}
 	if ( recField != 0)
-		recRef = (DNSRecordRef) (*pEnv)->GetIntField( pEnv, destObj, recField);
+		recRef = (DNSRecordRef) (*pEnv)->GetIntField( pEnv, pThis, recField);
 	if ( pContext == NULL || pContext->ServiceRef == NULL)
 		return kDNSServiceErr_BadParam;
 
@@ -595,31 +631,35 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_UpdateRecord( JNIE
 	return err;
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleRegistration_RemoveRecord( JNIEnv *pEnv, jobject pThis, 
-							jobject destObj, jint flags)
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSRecord_Remove( JNIEnv *pEnv, jobject pThis)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
-	jfieldID				contextField = (*pEnv)->GetFieldID( pEnv, cls, "fNativeContext", "I");
-	jclass					destCls = (*pEnv)->GetObjectClass( pEnv, destObj);
-	jfieldID				recField = (*pEnv)->GetFieldID( pEnv, destCls, "fRecord", "I");
+	jfieldID				ownerField = (*pEnv)->GetFieldID( pEnv, cls, "fOwner", "Lcom/apple/dnssd/AppleService;");
+	jfieldID				recField = (*pEnv)->GetFieldID( pEnv, cls, "fRecord", "I");
 	OpContext				*pContext = NULL;
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 	DNSRecordRef			recRef = NULL;
 
-	if ( contextField != 0)
-		pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, pThis, contextField);
+	if ( ownerField != 0)
+	{
+		jobject		ownerObj = (*pEnv)->GetObjectField( pEnv, pThis, ownerField);
+		jclass		ownerClass = (*pEnv)->GetObjectClass( pEnv, ownerObj);
+		jfieldID	contextField = (*pEnv)->GetFieldID( pEnv, ownerClass, "fNativeContext", "I");
+		if ( contextField != 0)
+			pContext = (OpContext*) (*pEnv)->GetIntField( pEnv, ownerObj, contextField);
+	}
 	if ( recField != 0)
-		recRef = (DNSRecordRef) (*pEnv)->GetIntField( pEnv, destObj, recField);
+		recRef = (DNSRecordRef) (*pEnv)->GetIntField( pEnv, pThis, recField);
 	if ( pContext == NULL || pContext->ServiceRef == NULL)
 		return kDNSServiceErr_BadParam;
 
-	err = DNSServiceRemoveRecord( pContext->ServiceRef, recRef, flags);
+	err = DNSServiceRemoveRecord( pContext->ServiceRef, recRef, 0);
 
 	return err;
 }
 
 
-static void CALLBACK_COMPAT	ServiceQueryReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex, 
+static void DNSSD_API	ServiceQueryReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex,
 								DNSServiceErrorType errorCode, const char *serviceName,
 								uint16_t rrtype, uint16_t rrclass, uint16_t rdlen,
 								const void *rdata, uint32_t ttl, void *context)
@@ -640,8 +680,8 @@ static void CALLBACK_COMPAT	ServiceQueryReply( DNSServiceRef sdRef _UNUSED, DNSS
 			memcpy( pBytes, rdata, rdlen);
 			(*pContext->Env)->ReleaseByteArrayElements( pContext->Env, rDataObj, pBytes, JNI_COMMIT);
 	
-			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback, 
-								pContext->JavaObj, flags, interfaceIndex, 
+			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, pContext->Callback,
+								pContext->JavaObj, flags, interfaceIndex,
 								(*pContext->Env)->NewStringUTF( pContext->Env, serviceName),
 								rrtype, rrclass, rDataObj, ttl);
 		}
@@ -651,7 +691,7 @@ static void CALLBACK_COMPAT	ServiceQueryReply( DNSServiceRef sdRef _UNUSED, DNSS
 	TeardownCallbackState();
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleQuery_CreateQuery( JNIEnv *pEnv, jobject pThis, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleQuery_CreateQuery( JNIEnv *pEnv, jobject pThis,
 							jint flags, jint ifIndex, jstring serviceName, jint rrtype, jint rrclass)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -660,7 +700,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleQuery_CreateQuery( JNIEnv *pEnv
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 
 	if ( contextField != 0)
-		pContext = NewContext( pEnv, pThis, "Lcom/apple/dnssd/QueryListener;", "queryAnswered", 
+		pContext = NewContext( pEnv, pThis, "queryAnswered",
 								"(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;II[BI)V");
 	else
 		err = kDNSServiceErr_BadParam;
@@ -669,7 +709,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleQuery_CreateQuery( JNIEnv *pEnv
 	{
 		const char	*servStr = SafeGetUTFChars( pEnv, serviceName);
 
-		err = DNSServiceQueryRecord( &pContext->ServiceRef, flags, ifIndex, servStr, 
+		err = DNSServiceQueryRecord( &pContext->ServiceRef, flags, ifIndex, servStr,
 									rrtype, rrclass, ServiceQueryReply, pContext);
 		if ( err == kDNSServiceErr_NoError)
 		{
@@ -685,7 +725,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleQuery_CreateQuery( JNIEnv *pEnv
 }
 
 
-static void CALLBACK_COMPAT	DomainEnumReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex, 
+static void DNSSD_API	DomainEnumReply( DNSServiceRef sdRef _UNUSED, DNSServiceFlags flags, uint32_t interfaceIndex,
 								DNSServiceErrorType errorCode, const char *replyDomain, void *context)
 {
 	OpContext		*pContext = (OpContext*) context;
@@ -696,9 +736,9 @@ static void CALLBACK_COMPAT	DomainEnumReply( DNSServiceRef sdRef _UNUSED, DNSSer
 	{
 		if ( errorCode == kDNSServiceErr_NoError)
 		{
-			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj, 
-								( flags & kDNSServiceFlagsAdd) != 0 ? pContext->Callback : pContext->Callback2, 
-								pContext->JavaObj, flags, interfaceIndex, 
+			(*pContext->Env)->CallVoidMethod( pContext->Env, pContext->ClientObj,
+								( flags & kDNSServiceFlagsAdd) != 0 ? pContext->Callback : pContext->Callback2,
+								pContext->JavaObj, flags, interfaceIndex,
 								(*pContext->Env)->NewStringUTF( pContext->Env, replyDomain));
 		}
 		else
@@ -707,7 +747,7 @@ static void CALLBACK_COMPAT	DomainEnumReply( DNSServiceRef sdRef _UNUSED, DNSSer
 	TeardownCallbackState();
 }
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDomainEnum_BeginEnum( JNIEnv *pEnv, jobject pThis, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDomainEnum_BeginEnum( JNIEnv *pEnv, jobject pThis,
 							jint flags, jint ifIndex)
 {
 	jclass					cls = (*pEnv)->GetObjectClass( pEnv, pThis);
@@ -716,18 +756,18 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDomainEnum_BeginEnum( JNIEnv *p
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
 
 	if ( contextField != 0)
-		pContext = NewContext( pEnv, pThis, "Lcom/apple/dnssd/DomainListener;", "domainFound", 
+		pContext = NewContext( pEnv, pThis, "domainFound",
 								"(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;)V");
 	else
 		err = kDNSServiceErr_BadParam;
 
 	if ( pContext != NULL)
 	{
-		pContext->Callback2 = (*pEnv)->GetMethodID( pEnv, 
-								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj), 
+		pContext->Callback2 = (*pEnv)->GetMethodID( pEnv,
+								(*pEnv)->GetObjectClass( pEnv, pContext->ClientObj),
 								"domainLost", "(Lcom/apple/dnssd/DNSSDService;IILjava/lang/String;)V");
 
-		err = DNSServiceEnumerateDomains( &pContext->ServiceRef, flags, ifIndex, 
+		err = DNSServiceEnumerateDomains( &pContext->ServiceRef, flags, ifIndex,
 											DomainEnumReply, pContext);
 		if ( err == kDNSServiceErr_NoError)
 		{
@@ -741,7 +781,7 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDomainEnum_BeginEnum( JNIEnv *p
 }
 
 
-JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_ConstructName( JNIEnv *pEnv, jobject pThis _UNUSED, 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_ConstructName( JNIEnv *pEnv, jobject pThis _UNUSED,
 							jstring serviceName, jstring regtype, jstring domain, jobjectArray pOut)
 {
 	DNSServiceErrorType		err = kDNSServiceErr_NoError;
@@ -765,8 +805,8 @@ JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_ConstructName( JNIEnv *pE
 	return err;
 }
 
-JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleDNSSD_ReconfirmRecord( JNIEnv *pEnv, jobject pThis _UNUSED, 
-							jint flags, jint ifIndex, jstring fullName, 
+JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleDNSSD_ReconfirmRecord( JNIEnv *pEnv, jobject pThis _UNUSED,
+							jint flags, jint ifIndex, jstring fullName,
 							jint rrtype, jint rrclass, jbyteArray rdata)
 {
 	jbyte					*pBytes;
@@ -784,5 +824,141 @@ JNIEXPORT void JNICALL Java_com_apple_dnssd_AppleDNSSD_ReconfirmRecord( JNIEnv *
 	SafeReleaseUTFChars( pEnv, fullName, nameStr);
 }
 
+#define LOCAL_ONLY_NAME "loo"
+
+JNIEXPORT jstring JNICALL Java_com_apple_dnssd_AppleDNSSD_GetNameForIfIndex( JNIEnv *pEnv, jobject pThis _UNUSED,
+							jint ifIndex)
+{
+	char					*p = LOCAL_ONLY_NAME, nameBuff[IF_NAMESIZE];
+
+	if (ifIndex != kDNSServiceInterfaceIndexLocalOnly)
+		p = if_indextoname( ifIndex, nameBuff );
+
+	return (*pEnv)->NewStringUTF( pEnv, p);
+}
 
 
+JNIEXPORT jint JNICALL Java_com_apple_dnssd_AppleDNSSD_GetIfIndexForName( JNIEnv *pEnv, jobject pThis _UNUSED,
+							jstring ifName)
+{
+	uint32_t				ifIndex = kDNSServiceInterfaceIndexLocalOnly;
+	const char				*nameStr = SafeGetUTFChars( pEnv, ifName);
+
+	if (strcmp(nameStr, LOCAL_ONLY_NAME))
+		ifIndex = if_nametoindex( nameStr);
+
+	SafeReleaseUTFChars( pEnv, ifName, nameStr);
+
+	return ifIndex;
+}
+
+
+#if defined(_WIN32)
+static char*
+if_indextoname( DWORD ifIndex, char * nameBuff)
+{
+	PIP_ADAPTER_INFO	pAdapterInfo = NULL;
+	PIP_ADAPTER_INFO	pAdapter = NULL;
+	DWORD				dwRetVal = 0;
+	char			*	ifName = NULL;
+	ULONG				ulOutBufLen = 0;
+	
+	if (GetAdaptersInfo( NULL, &ulOutBufLen) != ERROR_BUFFER_OVERFLOW)
+	{
+		goto exit;
+	}
+
+	pAdapterInfo = (IP_ADAPTER_INFO *) malloc(ulOutBufLen); 
+
+	if (pAdapterInfo == NULL)
+	{
+		goto exit;
+	}
+
+	dwRetVal = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen );
+	
+	if (dwRetVal != NO_ERROR)
+	{
+		goto exit;
+	}
+
+	pAdapter = pAdapterInfo;
+	while (pAdapter)
+	{
+		if (pAdapter->Index == ifIndex)
+		{
+			// It would be better if we passed in the length of nameBuff to this
+			// function, so we would have absolute certainty that no buffer
+			// overflows would occur.  Buffer overflows *shouldn't* occur because
+			// nameBuff is of size MAX_ADAPTER_NAME_LENGTH.
+			strcpy( nameBuff, pAdapter->AdapterName );
+			ifName = nameBuff;
+			break;
+		}
+  
+		pAdapter = pAdapter->Next;
+	}
+
+exit:
+
+	if (pAdapterInfo != NULL)
+	{
+		free( pAdapterInfo );
+		pAdapterInfo = NULL;
+	}
+
+	return ifName;
+}
+
+
+static DWORD
+if_nametoindex( const char * nameStr )
+{
+	PIP_ADAPTER_INFO	pAdapterInfo = NULL;
+	PIP_ADAPTER_INFO	pAdapter = NULL;
+	DWORD				dwRetVal = 0;
+	DWORD				ifIndex = 0;
+	ULONG				ulOutBufLen = 0;
+
+	if (GetAdaptersInfo( NULL, &ulOutBufLen) != ERROR_BUFFER_OVERFLOW)
+	{
+		goto exit;
+	}
+
+	pAdapterInfo = (IP_ADAPTER_INFO *) malloc(ulOutBufLen); 
+
+	if (pAdapterInfo == NULL)
+	{
+		goto exit;
+	}
+
+	dwRetVal = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen );
+	
+	if (dwRetVal != NO_ERROR)
+	{
+		goto exit;
+	}
+
+	pAdapter = pAdapterInfo;
+	while (pAdapter)
+	{
+		if (strcmp(pAdapter->AdapterName, nameStr) == 0)
+		{
+			ifIndex = pAdapter->Index;
+			break;
+		}
+  
+		pAdapter = pAdapter->Next;
+	}
+
+exit:
+
+	if (pAdapterInfo != NULL)
+	{
+		free( pAdapterInfo );
+		pAdapterInfo = NULL;
+	}
+
+	return ifIndex;
+}
+#endif
